# 数据合并
from abc import ABC, abstractmethod
from re import findall, match, sub


# 合并数据的基本行为
class DataMerge(ABC):
    @abstractmethod
    def merge(self, arg1, arg2):
        pass

    def merges(self, h):
        pass


# 合并数据
class DefaultMerge(DataMerge):
    priority = {
        "/": 2,
        "%": 2,
        "*": 2,
        "+": 1,
        "-": 1
    }

    def __init__(self, rules_dict):
        if isinstance(rules_dict, dict):
            self.rules = []
            self.exper = []
            self.key_name = []
            for key in rules_dict.keys():
                self.rules.append(key.split("."))
                self.exper.append(rules_dict[key]["exper"])
                self.key_name.append(rules_dict[key]["name"])
        else:
            raise Exception()

    # 中缀表达式转后缀表达式
    def suffix_exper(self, exp):
        # 关键字存储栈
        warehouse = []
        suffix_expr = []
        i = 0
        while i < len(exp):
            exp_ = exp[i]
            if exp_ in "(":
                warehouse.append("(")
                i += 1
            elif exp_ in ")":
                while len(warehouse) > 0 and warehouse[-1] != "(":
                    suffix_expr.append(warehouse.pop())
                warehouse.pop()
                i += 1
            elif exp_ in " ":
                i += 1
            else:
                parse_val = ""
                j = i
                while j < len(exp) and exp[j] not in "() ":
                    parse_val += exp[j]
                    j += 1
                i = j

                if parse_val in self.priority:
                    warehouse.append(parse_val)
                else:
                    suffix_expr.append(parse_val)
        while len(warehouse) > 0:
            suffix_expr.append(warehouse.pop())
        return suffix_expr

    # 后缀表达式计算结果
    def calc_exper(self, suffix_exper):
        suffix = []
        for exp in suffix_exper:
            if exp in self.priority:
                a = float(suffix.pop())
                b = float(suffix.pop())
                if exp in "*":
                    suffix.append(b * a)
                elif exp in "/":
                    suffix.append(b / a)
                elif exp in "+":
                    suffix.append(b + a)
                elif exp in "-":
                    suffix.append(b - a)
                elif exp == "%":
                    suffix.append(b / a)
            else:
                suffix.append(exp)
        return suffix.pop()

    # 深度优先遍历context，通过layer_num去匹配exper和rules，把匹配后的
    # layer_num: 记录当前变量处于第几层。
    # layer_key_name: 记录深度上的key名
    def parse(self, context, rules, exper, origin):
        if isinstance(context, dict):
            key_name_stack = list(reversed([key for key in context.keys()]))
            # 深度遍历context字典
            obj_stack = [context[key] for key in list(reversed([key for key in context.keys()]))]
            # 维护变量对应的深度
            layer_stack = [0 for key in range(len(obj_stack))]
            # 维护一个深度上的变量名
            layer_key_name = []
            last_layer = -1
            while len(key_name_stack) > 0:
                key_name = key_name_stack.pop()
                pack_obj = obj_stack.pop()
                layer_num = layer_stack.pop()

                # 维护一个变量名栈，用来记录一个深度上的变量
                if layer_num > last_layer:
                    layer_key_name.append(key_name)
                else:
                    while last_layer >= layer_num:
                        layer_key_name.pop()
                        last_layer -= 1
                    layer_key_name.append(key_name)

                last_layer = layer_num
                if layer_num == 4:
                    print(layer_key_name)

                # 规则匹配
                if layer_num < len(rules) and layer_num == len(rules) - 1:
                    all_match = True
                    for i in range(layer_num):
                        key_name = layer_key_name[i]
                        all_match &= match(rules[i] + "$", key_name) is not None

                    if all_match:
                        exper_copy = str(exper)
                        # 搜索变量
                        exper_ = findall(r"\${([a-zA-Z0-9-.\u4e00-\u9fff]+)}", exper_copy)
                        for exp in exper_:
                            origin_val = pack_obj
                            exps = exp.split(".")
                            try:
                                for origin_name in exps:
                                    origin_val = origin_val[origin_name]
                            except KeyError:
                                exper_copy = None
                                break
                            # 把变量替换为常量
                            exper_copy = sub(r"\${" + exp + "}", str(origin_val), exper_copy)
                        if not exper_copy is None:
                            pack_obj[origin] = self.calc_exper(self.suffix_exper(exper_copy))

                # 深度优先
                if isinstance(pack_obj, dict):
                    items = list(reversed([item for item in pack_obj.keys()]))
                    for item in items:
                        key_name_stack.append(item)
                        obj_stack.append(pack_obj[item])
                        layer_stack.append(layer_num + 1)
        else:
            raise Exception()

    # TODO:合并有问题
    def merge(self, arg1, arg2):
        if isinstance(arg1, dict) and isinstance(arg2, dict):
            variable_stack = [item for item in arg2.keys()]
            obj2_stack = [arg2 for i in arg2.keys()]
            obj1_stack = [arg1 for i in arg2.keys()]
            while len(variable_stack) > 0:
                key = variable_stack.pop()
                obj1 = obj1_stack.pop()
                obj2 = obj2_stack.pop()
                if key not in obj1:
                    obj1[key] = obj2[key]
                if not isinstance(obj1[key], dict) and not isinstance(obj2[key], dict):
                    obj1[key] = obj2[key]
                else:
                    for item in obj2[key].keys():
                        obj2_stack.append(obj2[key])
                        obj1_stack.append(obj1[key])
                        variable_stack.append(item)
            return arg1
        else:
            raise Exception()

    # 合并数据
    def merges(self, h):
        if isinstance(h, list) and all(isinstance(item, dict) for item in h):
            if len(h) >= 2:
                while len(h) >= 2:
                    h.append(self.merge(h.pop(), h.pop()))
                merge_result = h.pop()
                for i in range(len(self.rules)):
                    self.parse(merge_result, self.rules[i], self.exper[i], str(self.key_name[i]))
                return merge_result
            else:
                raise Exception("数据长度必须大于等于2，不然合并有啥意义？")
        else:
            raise Exception()
